@@ -1,24 +1,21 @@
 /*
  * File loading code for Mini-XML, a small XML file parsing library.
  *
- * Copyright 2003-2017 by Michael R Sweet.
+ * https://www.msweet.org/mxml
  *
- * These coded instructions, statements, and computer programs are the
- * property of Michael R Sweet and are protected by Federal copyright
- * law.  Distribution and use rights are outlined in the file "COPYING"
- * which should have been included with this file.  If this file is
- * missing or damaged, see the license at:
+ * Copyright © 2003-2019 by Michael R Sweet.
  *
- *     https://michaelrsweet.github.io/mxml
+ * Licensed under Apache License v2.0.  See the file "LICENSE" for more
+ * information.
  */
 
 /*
  * Include necessary headers...
  */
 
-#ifndef WIN32
+#ifndef _WIN32
 #  include <unistd.h>
-#endif /* !WIN32 */
+#endif /* !_WIN32 */
 #include "mxml-private.h"
 
 
@@ -58,42 +55,26 @@ typedef struct _mxml_fdbuf_s		/**** File descriptor buffer ****/
  * Local functions...
  */
 
-static int		mxml_add_char(int ch, char **ptr, char **buffer,
-			              int *bufsize);
+static int		mxml_add_char(int ch, char **ptr, char **buffer, int *bufsize);
 static int		mxml_fd_getc(void *p, int *encoding);
 static int		mxml_fd_putc(int ch, void *p);
 static int		mxml_fd_read(_mxml_fdbuf_t *buf);
 static int		mxml_fd_write(_mxml_fdbuf_t *buf);
 static int		mxml_file_getc(void *p, int *encoding);
 static int		mxml_file_putc(int ch, void *p);
-static int		mxml_get_entity(mxml_node_t *parent, void *p,
-			                int *encoding,
-					_mxml_getc_cb_t getc_cb);
+static int		mxml_get_entity(mxml_node_t *parent, void *p, int *encoding, _mxml_getc_cb_t getc_cb, int *line);
 static inline int	mxml_isspace(int ch)
 			{
-			  return (ch == ' ' || ch == '\t' || ch == '\r' ||
-			          ch == '\n');
+			  return (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n');
 			}
-static mxml_node_t	*mxml_load_data(mxml_node_t *top, void *p,
-			                mxml_load_cb_t cb,
-			                _mxml_getc_cb_t getc_cb,
-                                        mxml_sax_cb_t sax_cb, void *sax_data);
-static int		mxml_parse_element(mxml_node_t *node, void *p,
-			                   int *encoding,
-					   _mxml_getc_cb_t getc_cb);
+static mxml_node_t	*mxml_load_data(mxml_node_t *top, void *p, mxml_load_cb_t cb, _mxml_getc_cb_t getc_cb, mxml_sax_cb_t sax_cb, void *sax_data);
+static int		mxml_parse_element(mxml_node_t *node, void *p, int *encoding, _mxml_getc_cb_t getc_cb, int *line);
 static int		mxml_string_getc(void *p, int *encoding);
 static int		mxml_string_putc(int ch, void *p);
-static int		mxml_write_name(const char *s, void *p,
-					_mxml_putc_cb_t putc_cb);
-static int		mxml_write_node(mxml_node_t *node, void *p,
-			                mxml_save_cb_t cb, int col,
-					_mxml_putc_cb_t putc_cb,
-					_mxml_global_t *global);
-static int		mxml_write_string(const char *s, void *p,
-					  _mxml_putc_cb_t putc_cb);
-static int		mxml_write_ws(mxml_node_t *node, void *p,
-			              mxml_save_cb_t cb, int ws,
-				      int col, _mxml_putc_cb_t putc_cb);
+static int		mxml_write_name(const char *s, void *p, _mxml_putc_cb_t putc_cb);
+static int		mxml_write_node(mxml_node_t *node, void *p, mxml_save_cb_t cb, int col, _mxml_putc_cb_t putc_cb, _mxml_global_t *global);
+static int		mxml_write_string(const char *s, void *p, _mxml_putc_cb_t putc_cb);
+static int		mxml_write_ws(mxml_node_t *node, void *p, mxml_save_cb_t cb, int ws, int col, _mxml_putc_cb_t putc_cb);
 
 
 /*
@@ -106,6 +87,12 @@ static int		mxml_write_ws(mxml_node_t *node, void *p,
  * The constants @code MXML_INTEGER_CALLBACK@, @code MXML_OPAQUE_CALLBACK@,
  * @code MXML_REAL_CALLBACK@, and @code MXML_TEXT_CALLBACK@ are defined for
  * loading child (data) nodes of the specified type.
+ *
+ * Note: The most common programming error when using the Mini-XML library is
+ * to load an XML file using the @code MXML_TEXT_CALLBACK@, which returns inline
+ * text as a series of whitespace-delimited words, instead of using the
+ * @code MXML_OPAQUE_CALLBACK@ which returns the inline text as a single string
+ * (including whitespace).
  */
 
 mxml_node_t *				/* O - First node or @code NULL@ if the file could not be read. */
@@ -142,6 +129,12 @@ mxmlLoadFd(mxml_node_t    *top,		/* I - Top node */
  * The constants @code MXML_INTEGER_CALLBACK@, @code MXML_OPAQUE_CALLBACK@,
  * @code MXML_REAL_CALLBACK@, and @code MXML_TEXT_CALLBACK@ are defined for
  * loading child (data) nodes of the specified type.
+ *
+ * Note: The most common programming error when using the Mini-XML library is
+ * to load an XML file using the @code MXML_TEXT_CALLBACK@, which returns inline
+ * text as a series of whitespace-delimited words, instead of using the
+ * @code MXML_OPAQUE_CALLBACK@ which returns the inline text as a single string
+ * (including whitespace).
  */
 
 mxml_node_t *				/* O - First node or @code NULL@ if the file could not be read. */
@@ -167,6 +160,12 @@ mxmlLoadFile(mxml_node_t    *top,	/* I - Top node */
  * The constants @code MXML_INTEGER_CALLBACK@, @code MXML_OPAQUE_CALLBACK@,
  * @code MXML_REAL_CALLBACK@, and @code MXML_TEXT_CALLBACK@ are defined for
  * loading child (data) nodes of the specified type.
+ *
+ * Note: The most common programming error when using the Mini-XML library is
+ * to load an XML file using the @code MXML_TEXT_CALLBACK@, which returns inline
+ * text as a series of whitespace-delimited words, instead of using the
+ * @code MXML_OPAQUE_CALLBACK@ which returns the inline text as a single string
+ * (including whitespace).
  */
 
 mxml_node_t *				/* O - First node or @code NULL@ if the string has errors. */
@@ -1300,8 +1299,9 @@ static int				/* O  - Character value or EOF on error */
 mxml_get_entity(mxml_node_t *parent,	/* I  - Parent node */
 		void        *p,		/* I  - Pointer to source */
 		int         *encoding,	/* IO - Character encoding */
-                int         (*getc_cb)(void *, int *))
+                int         (*getc_cb)(void *, int *),
 					/* I  - Get character function */
+                int         *line)	/* IO - Current line number */
 {
   int	ch;				/* Current character */
   char	entity[64],			/* Entity string */
@@ -1311,23 +1311,27 @@ mxml_get_entity(mxml_node_t *parent,	/* I  - Parent node */
   entptr = entity;
 
   while ((ch = (*getc_cb)(p, encoding)) != EOF)
+  {
     if (ch > 126 || (!isalnum(ch) && ch != '#'))
       break;
     else if (entptr < (entity + sizeof(entity) - 1))
       *entptr++ = ch;
     else
     {
-      mxml_error("Entity name too long under parent <%s>!",
-	         parent ? parent->value.element.name : "null");
+      mxml_error("Entity name too long under parent <%s> on line %d.", parent ? parent->value.element.name : "null", *line);
       break;
     }
+  }
 
   *entptr = '\0';
 
   if (ch != ';')
   {
-    mxml_error("Character entity \"%s\" not terminated under parent <%s>!",
-	       entity, parent ? parent->value.element.name : "null");
+    mxml_error("Character entity '%s' not terminated under parent <%s> on line %d.", entity, parent ? parent->value.element.name : "null", *line);
+
+    if (ch == '\n')
+      (*line)++;
+
     return (EOF);
   }
 
@@ -1339,13 +1343,11 @@ mxml_get_entity(mxml_node_t *parent,	/* I  - Parent node */
       ch = (int)strtol(entity + 1, NULL, 10);
   }
   else if ((ch = mxmlEntityGetValue(entity)) < 0)
-    mxml_error("Entity name \"%s;\" not supported under parent <%s>!",
-	       entity, parent ? parent->value.element.name : "null");
+    mxml_error("Entity name '%s;' not supported under parent <%s> on line %d.", entity, parent ? parent->value.element.name : "null", *line);
 
   if (mxml_bad_char(ch))
   {
-    mxml_error("Bad control character 0x%02x under parent <%s> not allowed by XML standard!",
-               ch, parent ? parent->value.element.name : "null");
+    mxml_error("Bad control character 0x%02x under parent <%s> on line %d not allowed by XML standard.", ch, parent ? parent->value.element.name : "null", *line);
     return (EOF);
   }
 
@@ -1369,7 +1371,8 @@ mxml_load_data(
   mxml_node_t	*node,			/* Current node */
 		*first,			/* First node added */
 		*parent;		/* Current parent node */
-  int		ch,			/* Character from file */
+  int		line = 1,		/* Current line number */
+		ch,			/* Character from file */
 		whitespace;		/* Non-zero if whitespace seen */
   char		*buffer,		/* String buffer */
 		*bufptr;		/* Pointer into buffer */
@@ -1413,7 +1416,19 @@ mxml_load_data(
   else
     type = MXML_IGNORE;
 
-  while ((ch = (*getc_cb)(p, &encoding)) != EOF)
+  if ((ch = (*getc_cb)(p, &encoding)) == EOF)
+  {
+    free(buffer);
+    return (NULL);
+  }
+  else if (ch != '<' && !top)
+  {
+    free(buffer);
+    mxml_error("XML does not start with '<' (saw '%c').", ch);
+    return (NULL);
+  }
+
+  do
   {
     if ((ch == '<' ||
          (mxml_isspace(ch) && type != MXML_OPAQUE && type != MXML_CUSTOM)) &&
@@ -1454,8 +1469,7 @@ mxml_load_data(
 
 	      if ((*global->custom_load_cb)(node, buffer))
 	      {
-	        mxml_error("Bad custom value '%s' in parent <%s>!",
-		           buffer, parent ? parent->value.element.name : "null");
+	        mxml_error("Bad custom value '%s' in parent <%s> on line %d.", buffer, parent ? parent->value.element.name : "null", line);
 		mxmlDelete(node);
 		node = NULL;
 	      }
@@ -1473,9 +1487,7 @@ mxml_load_data(
         * Bad integer/real number value...
 	*/
 
-        mxml_error("Bad %s value '%s' in parent <%s>!",
-	           type == MXML_INTEGER ? "integer" : "real", buffer,
-		   parent ? parent->value.element.name : "null");
+        mxml_error("Bad %s value '%s' in parent <%s> on line %d.", type == MXML_INTEGER ? "integer" : "real", buffer, parent ? parent->value.element.name : "null", line);
 	break;
       }
 
@@ -1488,8 +1500,7 @@ mxml_load_data(
 	* Print error and return...
 	*/
 
-	mxml_error("Unable to add value node of type %s to parent <%s>!",
-	           types[type], parent ? parent->value.element.name : "null");
+	mxml_error("Unable to add value node of type %s to parent <%s> on line %d.", types[type], parent ? parent->value.element.name : "null", line);
 	goto error;
       }
 
@@ -1507,6 +1518,9 @@ mxml_load_data(
     else if (mxml_isspace(ch) && type == MXML_TEXT)
       whitespace = 1;
 
+    if (ch == '\n')
+      line ++;
+
    /*
     * Add lone whitespace node if we have an element and existing
     * whitespace...
@@ -1542,6 +1556,7 @@ mxml_load_data(
       bufptr = buffer;
 
       while ((ch = (*getc_cb)(p, &encoding)) != EOF)
+      {
         if (mxml_isspace(ch) || ch == '>' || (ch == '/' && bufptr > buffer))
 	  break;
 	else if (ch == '<')
@@ -1551,7 +1566,7 @@ mxml_load_data(
 	}
 	else if (ch == '&')
 	{
-	  if ((ch = mxml_get_entity(parent, p, &encoding, getc_cb)) == EOF)
+	  if ((ch = mxml_get_entity(parent, p, &encoding, getc_cb, &line)) == EOF)
 	    goto error;
 
 	  if (mxml_add_char(ch, &bufptr, &buffer, &bufsize))
@@ -1566,6 +1581,10 @@ mxml_load_data(
 	         ((bufptr - buffer) == 8 && !strncmp(buffer, "![CDATA[", 8)))
 	  break;
 
+	if (ch == '\n')
+	  line ++;
+      }
+
       *bufptr = '\0';
 
       if (!strcmp(buffer, "!--"))
@@ -1581,6 +1600,9 @@ mxml_load_data(
 	    break;
 	  else if (mxml_add_char(ch, &bufptr, &buffer, &bufsize))
 	    goto error;
+
+	  if (ch == '\n')
+	    line ++;
 	}
 
        /*
@@ -1593,7 +1615,7 @@ mxml_load_data(
 	  * Print error and return...
 	  */
 
-	  mxml_error("Early EOF in comment node!");
+	  mxml_error("Early EOF in comment node on line %d.", line);
 	  goto error;
 	}
 
@@ -1610,8 +1632,7 @@ mxml_load_data(
 	  * There can only be one root element!
 	  */
 
-	  mxml_error("<%s> cannot be a second root node after <%s>",
-	             buffer, first->value.element.name);
+	  mxml_error("<%s> cannot be a second root node after <%s> on line %d.", buffer, first->value.element.name, line);
           goto error;
 	}
 
@@ -1621,8 +1642,7 @@ mxml_load_data(
 	  * Just print error for now...
 	  */
 
-	  mxml_error("Unable to add comment node to parent <%s>!",
-	             parent ? parent->value.element.name : "null");
+	  mxml_error("Unable to add comment node to parent <%s> on line %d.", parent ? parent->value.element.name : "null", line);
 	  break;
 	}
 
@@ -1656,6 +1676,9 @@ mxml_load_data(
 	  }
 	  else if (mxml_add_char(ch, &bufptr, &buffer, &bufsize))
 	    goto error;
+
+	  if (ch == '\n')
+	    line ++;
 	}
 
        /*
@@ -1668,7 +1691,7 @@ mxml_load_data(
 	  * Print error and return...
 	  */
 
-	  mxml_error("Early EOF in CDATA node!");
+	  mxml_error("Early EOF in CDATA node on line %d.", line);
 	  goto error;
 	}
 
@@ -1685,8 +1708,7 @@ mxml_load_data(
 	  * There can only be one root element!
 	  */
 
-	  mxml_error("<%s> cannot be a second root node after <%s>",
-	             buffer, first->value.element.name);
+	  mxml_error("<%s> cannot be a second root node after <%s> on line %d.", buffer, first->value.element.name, line);
           goto error;
 	}
 
@@ -1696,8 +1718,7 @@ mxml_load_data(
 	  * Print error and return...
 	  */
 
-	  mxml_error("Unable to add CDATA node to parent <%s>!",
-	             parent ? parent->value.element.name : "null");
+	  mxml_error("Unable to add CDATA node to parent <%s> on line %d.", parent ? parent->value.element.name : "null", line);
 	  goto error;
 	}
 
@@ -1724,6 +1745,9 @@ mxml_load_data(
 	    break;
 	  else if (mxml_add_char(ch, &bufptr, &buffer, &bufsize))
 	    goto error;
+
+	  if (ch == '\n')
+	    line ++;
 	}
 
        /*
@@ -1736,7 +1760,7 @@ mxml_load_data(
 	  * Print error and return...
 	  */
 
-	  mxml_error("Early EOF in processing instruction node!");
+	  mxml_error("Early EOF in processing instruction node on line %d.", line);
 	  goto error;
 	}
 
@@ -1752,8 +1776,7 @@ mxml_load_data(
 	  * There can only be one root element!
 	  */
 
-	  mxml_error("<%s> cannot be a second root node after <%s>",
-	             buffer, first->value.element.name);
+	  mxml_error("<%s> cannot be a second root node after <%s> on line %d.", buffer, first->value.element.name, line);
           goto error;
 	}
 
@@ -1763,8 +1786,7 @@ mxml_load_data(
 	  * Print error and return...
 	  */
 
-	  mxml_error("Unable to add processing instruction node to parent <%s>!",
-	             parent ? parent->value.element.name : "null");
+	  mxml_error("Unable to add processing instruction node to parent <%s> on line %d.", parent ? parent->value.element.name : "null", line);
 	  goto error;
 	}
 
@@ -1772,7 +1794,7 @@ mxml_load_data(
         {
           (*sax_cb)(node, MXML_SAX_DIRECTIVE, sax_data);
 
-          if (!mxmlRelease(node))
+          if (strncmp(node->value.element.name, "?xml ", 5) && !mxmlRelease(node))
             node = NULL;
         }
 
@@ -1805,12 +1827,17 @@ mxml_load_data(
 	  else
 	  {
             if (ch == '&')
-	      if ((ch = mxml_get_entity(parent, p, &encoding, getc_cb)) == EOF)
+            {
+	      if ((ch = mxml_get_entity(parent, p, &encoding, getc_cb, &line)) == EOF)
 		goto error;
+            }
 
 	    if (mxml_add_char(ch, &bufptr, &buffer, &bufsize))
 	      goto error;
 	  }
+
+	  if (ch == '\n')
+	    line ++;
 	}
         while ((ch = (*getc_cb)(p, &encoding)) != EOF);
 
@@ -1824,7 +1851,7 @@ mxml_load_data(
 	  * Print error and return...
 	  */
 
-	  mxml_error("Early EOF in declaration node!");
+	  mxml_error("Early EOF in declaration node on line %d.", line);
 	  goto error;
 	}
 
@@ -1840,8 +1867,7 @@ mxml_load_data(
 	  * There can only be one root element!
 	  */
 
-	  mxml_error("<%s> cannot be a second root node after <%s>",
-	             buffer, first->value.element.name);
+	  mxml_error("<%s> cannot be a second root node after <%s> on line %d.", buffer, first->value.element.name, line);
           goto error;
 	}
 
@@ -1851,8 +1877,7 @@ mxml_load_data(
 	  * Print error and return...
 	  */
 
-	  mxml_error("Unable to add declaration node to parent <%s>!",
-	             parent ? parent->value.element.name : "null");
+	  mxml_error("Unable to add declaration node to parent <%s> on line %d.", parent ? parent->value.element.name : "null", line);
 	  goto error;
 	}
 
@@ -1892,8 +1917,7 @@ mxml_load_data(
 	  * Close tag doesn't match tree; print an error for now...
 	  */
 
-	  mxml_error("Mismatched close tag <%s> under parent <%s>!",
-	             buffer, parent ? parent->value.element.name : "(null)");
+	  mxml_error("Mismatched close tag <%s> under parent <%s> on line %d.", buffer, parent ? parent->value.element.name : "(null)", line);
           goto error;
 	}
 
@@ -1934,8 +1958,7 @@ mxml_load_data(
 	  * There can only be one root element!
 	  */
 
-	  mxml_error("<%s> cannot be a second root node after <%s>",
-	             buffer, first->value.element.name);
+	  mxml_error("<%s> cannot be a second root node after <%s> on line %d.", buffer, first->value.element.name, line);
           goto error;
 	}
 
@@ -1945,22 +1968,20 @@ mxml_load_data(
 	  * Just print error for now...
 	  */
 
-	  mxml_error("Unable to add element node to parent <%s>!",
-	             parent ? parent->value.element.name : "null");
+	  mxml_error("Unable to add element node to parent <%s> on line %d.", parent ? parent->value.element.name : "null", line);
 	  goto error;
 	}
 
         if (mxml_isspace(ch))
         {
-	  if ((ch = mxml_parse_element(node, p, &encoding, getc_cb)) == EOF)
+	  if ((ch = mxml_parse_element(node, p, &encoding, getc_cb, &line)) == EOF)
 	    goto error;
         }
         else if (ch == '/')
 	{
 	  if ((ch = (*getc_cb)(p, &encoding)) != '>')
 	  {
-	    mxml_error("Expected > but got '%c' instead for element <%s/>!",
-	               ch, buffer);
+	    mxml_error("Expected > but got '%c' instead for element <%s/> on line %d.", ch, buffer, line);
             mxmlDelete(node);
             goto error;
 	  }
@@ -2007,7 +2028,7 @@ mxml_load_data(
       * Add character entity to current buffer...
       */
 
-      if ((ch = mxml_get_entity(parent, p, &encoding, getc_cb)) == EOF)
+      if ((ch = mxml_get_entity(parent, p, &encoding, getc_cb, &line)) == EOF)
 	goto error;
 
       if (mxml_add_char(ch, &bufptr, &buffer, &bufsize))
@@ -2023,6 +2044,7 @@ mxml_load_data(
 	goto error;
     }
   }
+  while ((ch = (*getc_cb)(p, &encoding)) != EOF);
 
  /*
   * Free the string buffer - we don't need it anymore...
@@ -2043,9 +2065,7 @@ mxml_load_data(
 
     if (node != parent)
     {
-      mxml_error("Missing close tag </%s> under parent <%s>!",
-	         node->value.element.name,
-		 node->parent ? node->parent->value.element.name : "(null)");
+      mxml_error("Missing close tag </%s> under parent <%s> on line %d.", node->value.element.name, node->parent ? node->parent->value.element.name : "(null)", line);
 
       mxmlDelete(first);
 
@@ -2062,7 +2082,7 @@ mxml_load_data(
   * Common error return...
   */
 
-error:
+  error:
 
   mxmlDelete(first);
 
@@ -2081,7 +2101,8 @@ mxml_parse_element(
     mxml_node_t     *node,		/* I  - Element node */
     void            *p,			/* I  - Data to read from */
     int             *encoding,		/* IO - Encoding */
-    _mxml_getc_cb_t getc_cb)		/* I  - Data callback */
+    _mxml_getc_cb_t getc_cb,		/* I  - Data callback */
+    int             *line)		/* IO - Current line number */
 {
   int	ch,				/* Current character in file */
 	quote;				/* Quoting character */
@@ -2128,7 +2149,12 @@ mxml_parse_element(
     */
 
     if (mxml_isspace(ch))
+    {
+      if (ch == '\n')
+        (*line)++;
+
       continue;
+    }
 
    /*
     * Stop at /, ?, or >...
@@ -2144,8 +2170,7 @@ mxml_parse_element(
 
       if (quote != '>')
       {
-        mxml_error("Expected '>' after '%c' for element %s, but got '%c'!",
-	           ch, node->value.element.name, quote);
+        mxml_error("Expected '>' after '%c' for element %s, but got '%c' on line %d.", ch, node->value.element.name, quote, *line);
         goto error;
       }
 
@@ -2153,7 +2178,7 @@ mxml_parse_element(
     }
     else if (ch == '<')
     {
-      mxml_error("Bare < in element %s!", node->value.element.name);
+      mxml_error("Bare < in element %s on line %d.", node->value.element.name, *line);
       goto error;
     }
     else if (ch == '>')
@@ -2177,8 +2202,12 @@ mxml_parse_element(
       while ((ch = (*getc_cb)(p, encoding)) != EOF)
       {
         if (ch == '&')
-	  if ((ch = mxml_get_entity(node, p, encoding, getc_cb)) == EOF)
+        {
+	  if ((ch = mxml_get_entity(node, p, encoding, getc_cb, line)) == EOF)
 	    goto error;
+	}
+	else if (ch == '\n')
+	  (*line)++;
 
 	if (mxml_add_char(ch, &ptr, &name, &namesize))
 	  goto error;
@@ -2194,40 +2223,59 @@ mxml_parse_element(
       */
 
       while ((ch = (*getc_cb)(p, encoding)) != EOF)
+      {
 	if (mxml_isspace(ch) || ch == '=' || ch == '/' || ch == '>' ||
 	    ch == '?')
+	{
+	  if (ch == '\n')
+	    (*line)++;
           break;
+        }
 	else
 	{
           if (ch == '&')
-	    if ((ch = mxml_get_entity(node, p, encoding, getc_cb)) == EOF)
+          {
+	    if ((ch = mxml_get_entity(node, p, encoding, getc_cb, line)) == EOF)
 	      goto error;
+          }
 
 	  if (mxml_add_char(ch, &ptr, &name, &namesize))
 	    goto error;
 	}
+      }
     }
 
     *ptr = '\0';
 
     if (mxmlElementGetAttr(node, name))
+    {
+      mxml_error("Duplicate attribute '%s' in element %s on line %d.", name, node->value.element.name, name, *line);
       goto error;
+    }
 
     while (ch != EOF && mxml_isspace(ch))
+    {
       ch = (*getc_cb)(p, encoding);
 
+      if (ch == '\n')
+        (*line)++;
+    }
+
     if (ch == '=')
     {
      /*
       * Read the attribute value...
       */
 
-      while ((ch = (*getc_cb)(p, encoding)) != EOF && mxml_isspace(ch));
+      while ((ch = (*getc_cb)(p, encoding)) != EOF && mxml_isspace(ch))
+      {
+        if (ch == '\n')
+          (*line)++;
+      }
 
       if (ch == EOF)
       {
-        mxml_error("Missing value for attribute '%s' in element %s!",
-	           name, node->value.element.name);
+        mxml_error("Missing value for attribute '%s' in element %s on line %d.", name, node->value.element.name, *line);
         goto error;
       }
 
@@ -2241,17 +2289,25 @@ mxml_parse_element(
 	ptr   = value;
 
         while ((ch = (*getc_cb)(p, encoding)) != EOF)
+        {
 	  if (ch == quote)
+	  {
 	    break;
+	  }
 	  else
 	  {
 	    if (ch == '&')
-	      if ((ch = mxml_get_entity(node, p, encoding, getc_cb)) == EOF)
+	    {
+	      if ((ch = mxml_get_entity(node, p, encoding, getc_cb, line)) == EOF)
 	        goto error;
+	    }
+	    else if (ch == '\n')
+	      (*line)++;
 
 	    if (mxml_add_char(ch, &ptr, &value, &valsize))
 	      goto error;
 	  }
+	}
 
         *ptr = '\0';
       }
@@ -2265,17 +2321,26 @@ mxml_parse_element(
 	ptr      = value + 1;
 
 	while ((ch = (*getc_cb)(p, encoding)) != EOF)
+	{
 	  if (mxml_isspace(ch) || ch == '=' || ch == '/' || ch == '>')
+	  {
+	    if (ch == '\n')
+	      (*line)++;
+
             break;
+          }
 	  else
 	  {
 	    if (ch == '&')
-	      if ((ch = mxml_get_entity(node, p, encoding, getc_cb)) == EOF)
+	    {
+	      if ((ch = mxml_get_entity(node, p, encoding, getc_cb, line)) == EOF)
 	        goto error;
+	    }
 
 	    if (mxml_add_char(ch, &ptr, &value, &valsize))
 	      goto error;
 	  }
+	}
 
         *ptr = '\0';
       }
@@ -2288,8 +2353,7 @@ mxml_parse_element(
     }
     else
     {
-      mxml_error("Missing value for attribute '%s' in element %s!",
-	         name, node->value.element.name);
+      mxml_error("Missing value for attribute '%s' in element %s on line %d.", name, node->value.element.name, *line);
       goto error;
     }
 
@@ -2307,8 +2371,7 @@ mxml_parse_element(
 
       if (quote != '>')
       {
-        mxml_error("Expected '>' after '%c' for element %s, but got '%c'!",
-	           ch, node->value.element.name, quote);
+        mxml_error("Expected '>' after '%c' for element %s, but got '%c' on line %d.", ch, node->value.element.name, quote, *line);
         ch = EOF;
       }
 
@@ -2331,7 +2394,7 @@ mxml_parse_element(
   * Common error return point...
   */
 
-error:
+  error:
 
   free(name);
   free(value);
@@ -2705,7 +2768,7 @@ mxml_write_node(mxml_node_t     *node,	/* I - Node to write */
 		*next;			/* Next node */
   int		i,			/* Looping var */
 		width;			/* Width of attr + value */
-  mxml_attr_t	*attr;			/* Current attribute */
+  _mxml_attr_t	*attr;			/* Current attribute */
   char		s[255];			/* Temporary string */
 
 
@@ -2863,7 +2926,7 @@ mxml_write_node(mxml_node_t     *node,	/* I - Node to write */
 	      col ++;
 	  }
 
-	  sprintf(s, "%d", current->value.integer);
+	  snprintf(s, sizeof(s), "%d", current->value.integer);
 	  if (mxml_write_string(s, p, putc_cb) < 0)
 	    return (-1);
 
@@ -2893,7 +2956,7 @@ mxml_write_node(mxml_node_t     *node,	/* I - Node to write */
 	      col ++;
 	  }
 
-	  sprintf(s, "%f", current->value.real);
+	  snprintf(s, sizeof(s), "%f", current->value.real);
 	  if (mxml_write_string(s, p, putc_cb) < 0)
 	    return (-1);
 
@@ -2954,34 +3017,52 @@ mxml_write_node(mxml_node_t     *node,	/* I - Node to write */
 
     if ((next = current->child) == NULL)
     {
-      while ((next = current->next) == NULL)
+      if (current == node)
       {
-        if (current == node)
-          break;
-
        /*
-	* The ? and ! elements are special-cases and have no end tags...
-	*/
+        * Don't traverse to sibling node if we are at the "root" node...
+        */
 
-	current = current->parent;
+        next = NULL;
+      }
+      else
+      {
+       /*
+        * Try the next sibling, and continue traversing upwards as needed...
+        */
 
-	if (current->value.element.name[0] != '!' &&
-	    current->value.element.name[0] != '?')
+	while ((next = current->next) == NULL)
 	{
-	  col = mxml_write_ws(current, p, cb, MXML_WS_BEFORE_CLOSE, col, putc_cb);
+	  if (current == node || !current->parent)
+	    break;
 
-	  if ((*putc_cb)('<', p) < 0)
-	    return (-1);
-	  if ((*putc_cb)('/', p) < 0)
-	    return (-1);
-	  if (mxml_write_string(current->value.element.name, p, putc_cb) < 0)
-	    return (-1);
-	  if ((*putc_cb)('>', p) < 0)
-	    return (-1);
+	 /*
+	  * The ? and ! elements are special-cases and have no end tags...
+	  */
 
-	  col += strlen(current->value.element.name) + 3;
+	  current = current->parent;
 
-	  col = mxml_write_ws(current, p, cb, MXML_WS_AFTER_CLOSE, col, putc_cb);
+	  if (current->value.element.name[0] != '!' &&
+	      current->value.element.name[0] != '?')
+	  {
+	    col = mxml_write_ws(current, p, cb, MXML_WS_BEFORE_CLOSE, col, putc_cb);
+
+	    if ((*putc_cb)('<', p) < 0)
+	      return (-1);
+	    if ((*putc_cb)('/', p) < 0)
+	      return (-1);
+	    if (mxml_write_string(current->value.element.name, p, putc_cb) < 0)
+	      return (-1);
+	    if ((*putc_cb)('>', p) < 0)
+	      return (-1);
+
+	    col += strlen(current->value.element.name) + 3;
+
+	    col = mxml_write_ws(current, p, cb, MXML_WS_AFTER_CLOSE, col, putc_cb);
+	  }
+
+	  if (current == node)
+	    break;
 	}
       }
     }
