@@ -1,7 +1,8 @@
 /*
- * Configuration file for Mini-XML, a small XML file parsing library.
+ * Visual Studio configuration file for Mini-XML, a small XML file parsing
+ * library.
  *
- * Copyright 2003-2017 by Michael R Sweet.
+ * Copyright 2003-2019 by Michael R Sweet.
  *
  * These coded instructions, statements, and computer programs are the
  * property of Michael R Sweet and are protected by Federal copyright
@@ -20,7 +21,7 @@
  * warnings from their use...
  *
  * Then Microsoft decided that they should ignore this in VC2008 and use
- * yet another define (_CRT_SECURE_NO_WARNINGS) instead.  Bastards.
+ * yet another define (_CRT_SECURE_NO_WARNINGS) instead...
  */
 
 #define _CRT_SECURE_NO_DEPRECATE
@@ -62,7 +63,7 @@
  * Version number...
  */
 
-#define MXML_VERSION "Mini-XML v2.11"
+#define MXML_VERSION "Mini-XML v3.1"
 
 
 /*
@@ -112,6 +113,11 @@ extern char	*_mxml_strdup(const char *);
 #    define strdup _mxml_strdup
 #  endif /* !HAVE_STRDUP */
 
+#  ifndef HAVE_STRLCAT
+extern size_t	_mxml_strlcat(char *, const char *, size_t);
+#    define strlcat _mxml_strlcat
+#  endif /* !HAVE_STRLCAT */
+
 #  ifndef HAVE_STRLCPY
 extern size_t	_mxml_strlcpy(char *, const char *, size_t);
 #    define strlcpy _mxml_strlcpy
